import numpy as np

__all__ = ["roundflat"]


def roundflat(x, a_low=2.0, a_high=8.0, d_low=0.005, d_high=1.2):
    """Return the log probability of the round flat prior.

    The round flat prior is completely uninformative inside the interval bounds
    ``d_low`` and ``d_high`` while smoothly going to -inf for values outside.
    ``a_low`` and ``a_high`` specify how quickly the density falls at the boundaries.

    Args:
        x (float): A parameter value in [0, inf) for which to compute the log probability
        a_low (float): Steepness of the prior at the boundary ``d_low``.
        a_high (float): Steepness of the prior at the boundary ``d_high``.
        d_low (float): Lower boundary for which the log probability is -2.
        d_high (float): Upper boundary for which the log probability is -2.

    Returns:
        The log probability for x.
    """
    if x <= 0:
        return -np.inf
    return -2 * ((x / d_low) ** (-2 * a_low) + (x / d_high) ** (2 * a_high))



from collections import namedtuple
from decimal import Decimal
import itertools

import numpy as np
from scipy.optimize import minimize


__all__ = ["expected_ucb", "parse_timecontrol", "TimeControl", "TimeControlBag"]


def expected_ucb(res, n_random_starts=100, alpha=1.96, random_state=None):
    """ Compute the expected (pessimistic) optimum of the optimization result.

    This will compute `gp_mean + alpha * gp_se` and find the optimum of that function.

    Parameters
    ----------
    * `res`  [`OptimizeResult`, scipy object]:
        The optimization result returned by a `bask` minimizer.
    * `n_random_starts` [int, default=100]:
        The number of random starts for the minimization of the surrogate model.
    * `alpha` [float, default=1.96]:
        Number of standard errors to add to each point.
    * `random_state` [int, RandomState instance, or None (default)]:
        Set random state to something other than None for reproducible
        results.
    Returns
    -------
    * `x` [list]: location of the minimum.
    * `fun` [float]: the surrogate function value at the minimum.
    """

    def func(x):
        reg = res.models[-1]
        mu, std = reg.predict(x.reshape(1, -1), return_std=True)
        return (mu + alpha * std)[0]

    xs = [res.space.transform([res.x]).tolist()]
    if n_random_starts > 0:
        xs.extend(
            res.space.transform(
                res.space.rvs(n_random_starts, random_state=random_state)
            ).tolist()
        )

    best_x = None
    best_fun = np.inf

    for x0 in xs:
        r = minimize(func, x0=x0, bounds=[(0.0, 1.0)] * len(res.space.bounds))

        if r.fun < best_fun:
            best_x = r.x
            best_fun = r.fun
    return res.space.inverse_transform(best_x[None, :])[0], best_fun


def parse_timecontrol(tc_string):
    if "+" in tc_string:
        return tuple([Decimal(x) for x in tc_string.split("+")])
    return (Decimal(tc_string),)


TC = namedtuple("TimeControl", ["time", "increment"],)


class TimeControl(TC):
    @classmethod
    def from_string(cls, tc_string):
        tc = parse_timecontrol(tc_string)
        inc = Decimal("0.0") if len(tc) == 1 else tc[1]
        return cls(time=Decimal(tc[0]), increment=inc,)

    def __str__(self):
        if abs(self.increment) < 1e-10:
            return f"{self.time}"
        return f"{self.time}+{self.increment}"


def _latin_1d(n):
    return (np.random.uniform(0, 1, size=n) + np.arange(n)) / n


def _probabilistic_round(x):
    return int(np.floor(x + np.random.uniform()))


class TimeControlBag(object):
    def __init__(self, tcs, bag_size=10, p=None):
        self.bag = []
        self.tcs = tcs
        if p is not None:
            self.p = p
        else:
            self.p = np.ones(len(tcs)) / len(tcs)
        self.bag_size = bag_size

    def next_tc(self):
        if self.bag is None or len(self.bag) == 0:
            out = []
            for p in self.p:
                out.append(_probabilistic_round(p * self.bag_size))
            tmp_bag = []
            for o, tc in zip(out, self.tcs):
                tmp_bag.extend(itertools.product(_latin_1d(o), [tc]))
            sorted_bag = sorted(tmp_bag)
            self.bag = [x[1] for x in sorted_bag]
        return self.bag.pop()



"""The server worker, which reads results and schedules new jobs."""
import json
import logging
import os
import re
import sys
from datetime import datetime
from time import sleep

try:
    import joblib
except ImportError:
    joblib = None
import numpy as np
try:
    import pandas as pd
except ImportError:
    pd = None
import scipy.stats
from bask import Optimizer
from skopt.learning.gaussian_process.kernels import Matern, ConstantKernel
from skopt.space import space as skspace
from skopt.utils import normalize_dimensions, create_result
try:
    from sqlalchemy.orm import sessionmaker
except ImportError:
    sessionmaker = None

from tune.priors import roundflat
from tune.utils import expected_ucb
from tune.db_workers.dbmodels import (
    Base,
    SqlJob,
    SqlUCIParam,
    SqlResult,
    SqlTimeControl,
    SqlTune,
)
from tune.db_workers.utils import (
    get_session_maker,
    create_sqlalchemy_engine,
    TimeControl,
    simple_penta_to_score,
)
from tune.io import InitStrings

__all__ = ["TuningServer"]


class TuningServer(object):
    def __init__(self, experiment_path, dbconfig_path, **kwargs):
        self.logger = logging.getLogger("TuningServer")
        self.experiment_path = experiment_path
        if os.path.isfile(dbconfig_path):
            with open(dbconfig_path, "r") as config_file:
                config = config_file.read().replace("\n", "")
                self.logger.debug(f"Reading DB config:\n{config}")
                self.connect_params = json.loads(config)
        else:
            raise ValueError("No dbconfig file found at provided path")

        self.engine = create_sqlalchemy_engine(self.connect_params)
        Base.metadata.create_all(self.engine)
        sm = sessionmaker(bind=self.engine)
        self.sessionmaker = get_session_maker(sm)

        if os.path.isfile(experiment_path):
            with open(experiment_path, "r+") as experiment_file:
                exp = experiment_file.read().replace("\n", "")
                self.logger.debug(f"Reading experiment config:\n{exp}")
                self.experiment = json.loads(exp)
                self.logger.debug(f"self.experiment = \n{self.experiment}")
        else:
            raise ValueError("No experiment config file found at provided path")
        self.time_controls = [
            TimeControl.from_strings(*x) for x in self.experiment["time_controls"]
        ]
        self.rng = np.random.RandomState(self.experiment.get("random_seed", 123))
        self.setup_tuner()

        try:
            os.makedirs("experiments")
        except FileExistsError:
            pass
        # TODO: in principle after deleting all jobs from the database,
        #       this could be problematic:
        self.pos = None
        self.chain = None
        if "tune_id" in self.experiment:
            self.resume_tuning()

    def write_experiment_file(self):
        with open(self.experiment_path, "w") as experiment_file:
            experiment_file.write(json.dumps(self.experiment, indent=2))

    def save_state(self):
        path = os.path.join(
            "experiments", f"data_tuneid_{self.experiment['tune_id']}.npz"
        )
        np.savez_compressed(
            path, np.array(self.opt.gp.pos_), np.array(self.opt.gp.chain_)
        )

    def resume_tuning(self):
        path = os.path.join(
            "experiments", f"data_tuneid_{self.experiment['tune_id']}.npz"
        )
        if os.path.exists(path):
            data = np.load(path)
            self.opt.gp.pos_ = data["arr_0"]
            self.opt.gp.chain_ = data["arr_1"]

    def parse_dimensions(self, param_dict):
        def make_numeric(s):
            try:
                return int(s)
            except ValueError:
                try:
                    return float(s)
                except ValueError:
                    return s

        dimensions = []
        for s in param_dict.values():
            prior_str = re.findall(r"(\w+)\(", s)[0]
            prior_param_strings = re.findall(r"\((.*?)\)", s)[0].split(",")
            keys = [x.split("=")[0].strip() for x in prior_param_strings]
            vals = [make_numeric(x.split("=")[1].strip()) for x in prior_param_strings]
            dim = getattr(skspace, prior_str)(**dict(zip(keys, vals)))
            dimensions.append(dim)
        return dimensions

    def parse_priors(self, priors):
        if isinstance(priors, str):
            try:
                result = joblib.load(priors)
            except IOError:
                self.logger.error(
                    f"Priors could not be loaded from path {priors}. Terminating..."
                )
                sys.exit(1)
        else:
            result = []
            for i, p in enumerate(priors):
                prior_str = re.findall(r"(\w+)\(", p)[0]
                prior_param_strings = re.findall(r"\((.*?)\)", p)[0].split(",")
                keys = [x.split("=")[0].strip() for x in prior_param_strings]
                vals = [float(x.split("=")[1].strip()) for x in prior_param_strings]
                if prior_str == "roundflat":
                    prior = lambda x, keys=keys, vals=vals: roundflat(np.exp(x), **dict(zip(keys, vals))) + x
                else:
                    dist = getattr(scipy.stats, prior_str)(**dict(zip(keys, vals)))
                    if i == 0 or i == len(priors) - 1:
                        # The signal variance and the signal noise are in positive, sqrt domain
                        prior = (
                            lambda x, dist=dist: dist.logpdf(np.sqrt(np.exp(x)))
                            + x / 2.0
                            - np.log(2.0)
                        )
                    else:
                        # The lengthscale(s) are in positive domain
                        prior = lambda x, dist=dist: dist.logpdf(np.exp(x)) + x
                result.append(prior)
        return result

    def setup_tuner(self):
        self.tunecfg = self.experiment["tuner"]
        self.parameters = list(self.tunecfg["parameters"].keys())
        self.dimensions = self.parse_dimensions(self.tunecfg["parameters"])
        self.space = normalize_dimensions(self.dimensions)
        self.priors = self.parse_priors(self.tunecfg["priors"])

        self.kernel = ConstantKernel(
            constant_value=self.tunecfg.get("variance_value", 0.1 ** 2),
            constant_value_bounds=tuple(
                self.tunecfg.get("variance_bounds", (0.01 ** 2, 0.5 ** 2))
            ),
        ) * Matern(
            length_scale=self.tunecfg.get("length_scale_value", 0.3),
            length_scale_bounds=tuple(
                self.tunecfg.get("length_scale_bounds", (0.2, 0.8))
            ),
            nu=2.5,
        )
        self.opt = Optimizer(
            dimensions=self.dimensions,
            n_points=self.tunecfg.get("n_points", 1000),
            n_initial_points=self.tunecfg.get(
                "n_initial_points", 5 * len(self.dimensions)
            ),
            gp_kernel=self.kernel,
            gp_kwargs=dict(normalize_y=True),
            gp_priors=self.priors,
            acq_func=self.tunecfg.get("acq_func", "ts"),
            acq_func_kwargs=self.tunecfg.get(
                "acq_func_kwargs", None
            ),  # TODO: Check if this works for all parameters
            random_state=self.rng.randint(0, np.iinfo(np.int32).max),
        )

    def query_data(self, session, include_active=False):
        tune_id = self.experiment["tune_id"]
        # First check if samplesize was reached:
        sample_sizes = np.array(
            session.query(
                SqlResult.ww_count
                + SqlResult.wd_count
                + SqlResult.wl_count
                + SqlResult.dd_count
                + SqlResult.dl_count
                + SqlResult.ll_count
            )
            .join(SqlJob)
            .filter(SqlJob.active, SqlJob.tune_id == tune_id)
            .all()
        ).squeeze()
        samplesize_reached = False
        if np.all(sample_sizes >= self.experiment.get("minimum_samplesize", 16)):
            samplesize_reached = True

        q = session.query(SqlJob).filter(SqlJob.tune_id == tune_id)
        if not include_active:
            q = q.filter(SqlJob.active == False)  # noqa
        jobs = q.all()
        query = session.query(SqlUCIParam.job_id, SqlUCIParam.key, SqlUCIParam.value).join(SqlJob).filter(SqlJob.tune_id == tune_id)
        df = pd.read_sql(query.statement, query.session.bind)
        df['value'] = df['value'].astype(float)
        self.logger.debug(f"Data frame: {df.head()}")
        X = df.pivot(index="job_id", columns="key").sort_index().droplevel(0, axis=1)[self.parameters].values
        y = {tc: [] for tc in self.time_controls}
        for job in jobs:
            for result in job.results:
                tc = result.time_control.to_tuple()
                if tc not in self.time_controls:
                    continue
                draw_rate = float(result.time_control.draw_rate)
                counts = np.array(
                    [
                        result.ll_count,
                        result.dl_count,
                        result.wl_count + result.dd_count,
                        result.wd_count,
                        result.ww_count,
                    ]
                )
                score = simple_penta_to_score(
                    draw_rate=draw_rate, counts=counts,
                    prior_games=self.experiment.get("prior_games", 1),
                    prior_elo=self.experiment.get("prior_elo", 0.0)
                )
                y[tc].append(-score)
        return X, np.array(list(y.values())).mean(axis=0), samplesize_reached

    @staticmethod
    def change_engine_config(engine_config, params):
        init_strings = InitStrings(
            engine_config[0]["initStrings"]
        )  # TODO: allow tuning of different index
        for k, v in params.items():
            init_strings[k] = v

    def insert_jobs(self, session, new_x):
        # First set all active jobs to inactive:
        session.query(SqlJob).filter(
            SqlJob.active == True, SqlJob.tune_id == self.experiment["tune_id"]
        ).update(
            {"active": False}
        )  # noqa

        # Insert new job:
        job_dict = {
            "engine": self.experiment["engine"],
            "cutechess": self.experiment["cutechess"],
        }
        job_json = json.dumps(job_dict)
        job = SqlJob(
            minimum_version=self.experiment.get("minimum_version", 1),
            maximum_version=self.experiment.get("maximum_version", None),
            minimum_samplesize=self.experiment.get("minimum_samplesize", 16),
            config=job_json,
            engine1_exe=self.experiment.get("engine1_exe", "lc0"),
            engine1_nps=self.experiment["engine1_nps"],
            engine2_exe=self.experiment.get("engine2_exe", "sf"),
            engine2_nps=self.experiment["engine2_nps"],
            tune_id=self.experiment["tune_id"],
        )
        session.add(job)
        for i, tc in enumerate(self.time_controls):
            sql_tc = (
                session.query(SqlTimeControl)
                .filter(
                    SqlTimeControl.engine1_time == tc.engine1_time,
                    SqlTimeControl.engine1_increment == tc.engine1_increment,
                    SqlTimeControl.engine2_time == tc.engine2_time,
                    SqlTimeControl.engine2_increment == tc.engine2_increment,
                )
                .one_or_none()
            )
            if sql_tc is None:
                sql_tc = SqlTimeControl(
                    engine1_time=tc.engine1_time,
                    engine1_increment=tc.engine1_increment,
                    engine2_time=tc.engine2_time,
                    engine2_increment=tc.engine2_increment,
                )
                session.add(sql_tc)

            result = SqlResult(job=job, time_control=sql_tc)
            session.add(result)
        for k, v in zip(self.parameters, new_x):
            param = SqlUCIParam(key=k, value=str(v), job=job)
            session.add(param)

    def run(self):
        # 0. Before we run the main loop, do we need to initialize or resume?
        #    * Resume from files (in experiment folder)
        #    * Create tune entry in db if it does not exist yet

        if "tune_id" not in self.experiment:
            with self.sessionmaker() as session:
                tune = SqlTune(
                    weight=self.experiment.get("weight", 1.0),
                    description=self.experiment.get("description", None),
                )
                session.add(tune)
                session.flush()
                self.experiment["tune_id"] = tune.id
                self.write_experiment_file()
                new_x = self.opt.ask()
                # Alter engine json using Initstrings
                params = dict(zip(self.parameters, new_x))
                self.change_engine_config(self.experiment["engine"], params)
                self.insert_jobs(session, new_x)
                self.logger.info("New jobs committed to database.")
        while True:
            self.logger.debug("Begin querying for new data...")
            # Check if minimum sample size and minimum wait time are reached, then query data and update model:
            with self.sessionmaker() as session:
                X, y, samplesize_reached = self.query_data(session, include_active=True)
                self.logger.debug(
                    f"Queried the database for data and got (last 5):\n{X[-5:]}\n{y[-5:]}"
                )
                if len(X) == 0:
                    self.logger.info("There are no datapoints yet, start first job")
                    new_x = self.opt.ask()
                    # Alter engine json using Initstrings
                    params = dict(zip(self.parameters, new_x))
                    self.change_engine_config(self.experiment["engine"], params)
                    self.insert_jobs(session, new_x)
                    self.logger.info("New jobs committed to database.")
                    samplesize_reached = False

            if not samplesize_reached:
                sleep_seconds = self.experiment.get("sleep_time", 60)
                self.logger.debug(
                    f"Required sample size not yet reached. Sleeping {sleep_seconds} seconds."
                )
                sleep(sleep_seconds)
                continue

            # Tell optimizer about the new results:
            now = datetime.now()
            self.opt.tell(
                X.tolist(),
                y.tolist(),
                fit=True,
                replace=True,
                n_samples=self.tunecfg["n_samples"],
                gp_samples=self.tunecfg["gp_samples"],
                gp_burnin=self.tunecfg["gp_burnin"],
                progress=False,
            )
            later = datetime.now()
            difference = (later - now).total_seconds()
            self.logger.info(
                f"Calculating GP posterior and acquisition function finished in {difference}s"
            )
            self.logger.info(f"Current GP kernel:\n{self.opt.gp.kernel_}")
            if self.opt.gp.chain_ is not None:
                self.logger.debug("Saving position and chain")
                self.save_state()

            # Ask optimizer for new configuration and insert jobs:
            new_x = self.opt.ask()
            # Alter engine json using Initstrings
            params = dict(zip(self.parameters, new_x))
            self.change_engine_config(self.experiment["engine"], params)
            with self.sessionmaker() as session:
                self.insert_jobs(session, new_x)
            self.logger.info("New jobs committed to database.")
            sleep(self.experiment.get("sleep_time", 60))

            result_object = create_result(
                Xi=X.tolist(), yi=y.tolist(), space=self.opt.space, models=[self.opt.gp]
            )
            try:
                opt_x, opt_y = expected_ucb(result_object)
                self.logger.info(
                    f"Current optimum: {dict(zip(self.parameters, np.around(opt_x,4)))}"
                )
            except ValueError:
                self.logger.info(f"Current optimum: None (optimizer errored out :( )")

    def deactivate(self):
        raise NotImplementedError

    def reactivate(self):
        raise NotImplementedError



"""Console script for chess_tuning_tools."""
from datetime import datetime
import json
import logging
import os
import pathlib
import sys
import time

from atomicwrites import AtomicWriter
from bask.optimizer import Optimizer
import click
import matplotlib.pyplot as plt
import numpy as np
from skopt.utils import create_result


from tune.db_workers import TuningServer, TuningClient
from tune.local import run_match, parse_experiment_result, reduce_ranges
from tune.io import prepare_engines_json, load_tuning_config, write_engines_json
from tune.plots import plot_objective
from tune.utils import expected_ucb


@click.group()
def cli():
    pass


@cli.command()
@click.option(
    "--verbose", "-v", is_flag=True, default=False, help="Turn on debug output."
)
@click.option("--logfile", default=None, help="Path to where the log is saved to.")
@click.option(
    "--terminate-after", default=0, help="Terminate the client after x minutes."
)
@click.option(
    "--clientconfig", default=None, help="Path to the client configuration file."
)
@click.argument("dbconfig")
def run_client(verbose, logfile, terminate_after, clientconfig, dbconfig):
    """ Run the client to generate games for distributed tuning.

    In order to connect to the database you need to provide a valid DBCONFIG
    json file. It contains the necessary parameters to connect to the database
    where it can fetch jobs and store results.
    """
    log_level = logging.DEBUG if verbose else logging.INFO
    logging.basicConfig(
        level=log_level,
        filename=logfile,
        format="%(asctime)s %(levelname)-8s %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )
    tc = TuningClient(
        dbconfig_path=dbconfig,
        terminate_after=terminate_after,
        clientconfig=clientconfig,
    )
    tc.run()


@cli.command()
@click.option(
    "--verbose", "-v", is_flag=True, default=False, help="Turn on debug output."
)
@click.option("--logfile", default=None, help="Path to where the log is saved to.")
@click.argument("command")
@click.argument("experiment_file")
@click.argument("dbconfig")
def run_server(verbose, logfile, command, experiment_file, dbconfig):
    """Run the tuning server for a given EXPERIMENT_FILE (json).

    To connect to the database you also need to provide a DBCONFIG json file.

    \b
    You can choose from these COMMANDs:
     * run: Starts the server.
     * deactivate: Deactivates all active jobs of the given experiment.
     * reactivate: Reactivates all recent jobs for which sample size is not reached yet.
    """
    log_level = logging.DEBUG if verbose else logging.INFO
    logging.basicConfig(
        level=log_level,
        filename=logfile,
        format="%(asctime)s %(levelname)-8s %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )
    tc = TuningServer(experiment_path=experiment_file, dbconfig_path=dbconfig)
    if command == "run":
        tc.run()
    elif command == "deactivate":
        tc.deactivate()
    elif command == "reactivate":
        tc.reactivate()
    else:
        raise ValueError(f"Command {command} is not recognized. Terminating...")


@cli.command()
@click.option(
    "-c",
    "--tuning-config",
    help="json file containing the tuning configuration.",
    required=True,
    type=click.File("r"),
)
@click.option(
    "-a",
    "--acq-function",
    default="mes",
    help="Acquisition function to use for selecting points to try. "
    "Can be {mes, pvrs, ei, ts, vr}.",
    show_default=True,
)
@click.option(
    "--acq-function-samples",
    default=1,
    help="How many GP samples to average the acquisition function over. "
    "More samples will slow down the computation time, but might give more "
    "stable tuning results. Less samples on the other hand cause more exploration "
    "which could help avoid the tuning to get stuck.",
    show_default=True,
)
@click.option(
    "-d",
    "--data-path",
    default="data.npz",
    help="Save the evaluated points to this file.",
    type=click.Path(exists=False),
    show_default=True,
)
@click.option(
    "--gp-burnin",
    default=5,
    type=int,
    help="Number of samples to discard before sampling model parameters. "
    "This is used during tuning and few samples suffice.",
    show_default=True,
)
@click.option(
    "--gp-samples",
    default=300,
    type=int,
    help="Number of model parameters to sample for the model. "
    "This is used during tuning and it should be a multiple of 100.",
    show_default=True,
)
@click.option(
    "--gp-initial-burnin",
    default=100,
    type=int,
    help="Number of samples to discard before starting to sample the initial model "
    "parameters. This is only used when resuming or for the first model.",
    show_default=True,
)
@click.option(
    "--gp-initial-samples",
    default=300,
    type=int,
    help="Number of model parameters to sample for the initial model. "
    "This is only used when resuming or for the first model. "
    "Should be a multiple of 100.",
    show_default=True,
)
@click.option(
    "-l",
    "--logfile",
    default="log.txt",
    help="Path to log file.",
    type=click.Path(exists=False),
    show_default=True,
)
@click.option(
    "--n-initial-points",
    default=30,
    help="Size of initial dense set of points to try.",
    show_default=True,
)
@click.option(
    "--n-points",
    default=500,
    help="The number of random points to consider as possible next point. "
    "Less points reduce the computation time of the tuner, but reduce "
    "the coverage of the space.",
    show_default=True,
)
@click.option(
    "--plot-every",
    default=0,
    help="Plot the current optimization landscape every n-th iteration. "
    "Set to 0 to turn it off.",
    show_default=True,
)
@click.option(
    "--plot-path",
    default="plots",
    help="Path to the directory to which the tuner will output plots.",
    show_default=True,
)
@click.option(
    "--random-seed",
    default=0,
    help="Number to seed all internally used random generators.",
    show_default=True,
)
@click.option(
    "--result-every",
    default=5,
    help="Output the actual current optimum every n-th iteration."
    "The further you are in the tuning process, the longer this will take to "
    "compute. Consider increasing this number, if you do not need the output "
    "that often. Set to 0 to turn it off.",
    show_default=True,
)
@click.option(
    "--resume/--no-resume",
    default=True,
    help="Let the optimizer resume, if it finds points it can use.",
    show_default=True,
)
@click.option(
    "--verbose", "-v", is_flag=True, default=False, help="Turn on debug output."
)
def local(
    tuning_config,
    acq_function="mes",
    acq_function_samples=1,
    data_path=None,
    gp_burnin=5,
    gp_samples=300,
    gp_initial_burnin=100,
    gp_initial_samples=300,
    logfile="log.txt",
    n_initial_points=30,
    n_points=500,
    plot_every=5,
    plot_path="plots",
    random_seed=0,
    result_every=5,
    resume=True,
    verbose=False,
):
    """Run a local tune.

    Parameters defined in the `tuning_config` file always take precedence.
    """
    json_dict = json.load(tuning_config)
    settings, commands, fixed_params, param_ranges = load_tuning_config(json_dict)
    log_level = logging.DEBUG if verbose else logging.INFO
    log_format = logging.Formatter("%(asctime)s %(levelname)-8s %(message)s")
    root_logger = logging.getLogger()
    root_logger.setLevel(log_level)
    file_logger = logging.FileHandler(settings.get("logfile", logfile))
    file_logger.setFormatter(log_format)
    root_logger.addHandler(file_logger)
    console_logger = logging.StreamHandler(sys.stdout)
    console_logger.setFormatter(log_format)
    root_logger.addHandler(console_logger)
    logging.debug(f"Got the following tuning settings:\n{json_dict}")

    # 1. Create seed sequence
    ss = np.random.SeedSequence(settings.get("random_seed", random_seed))
    # 2. Create kernel
    # 3. Create optimizer
    random_state = np.random.RandomState(np.random.MT19937(ss.spawn(1)[0]))
    opt = Optimizer(
        dimensions=list(param_ranges.values()),
        n_points=settings.get("n_points", n_points),
        n_initial_points=settings.get("n_initial_points", n_initial_points),
        # gp_kernel=kernel,  # TODO: Let user pass in different kernels
        gp_kwargs=dict(normalize_y=True),
        # gp_priors=priors,  # TODO: Let user pass in priors
        acq_func=settings.get("acq_function", acq_function),
        acq_func_kwargs=dict(alpha="inf", n_thompson=20),
        random_state=random_state,
    )
    X = []
    y = []
    noise = []
    iteration = 0

    # 3.1 Resume from existing data:
    if data_path is None:
        data_path = "data.npz"
    if resume:
        path = pathlib.Path(data_path)
        if path.exists():
            importa = np.load(path)
            X = importa["arr_0"].tolist()
            y = importa["arr_1"].tolist()
            noise = importa["arr_2"].tolist()
            if len(X[0]) != opt.space.n_dims:
                logging.error(
                    "The number of parameters are not matching the number of dimensions. "
                    "Rename the existing data file or ensure that the parameter ranges are correct."
                )
                sys.exit(1)
            reduction_needed, X_reduced, y_reduced, noise_reduced = reduce_ranges(
                X, y, noise, opt.space
            )
            if reduction_needed:
                backup_path = path.parent / (
                    path.stem + f"_backup_{int(time.time())}" + path.suffix
                )
                logging.warning(
                    f"The parameter ranges are smaller than the existing data. "
                    f"Some points will have to be discarded. "
                    f"The original {iteration} data points will be saved to {backup_path}"
                )
                np.savez_compressed(
                    backup_path, np.array(X), np.array(y), np.array(noise)
                )
                X = X_reduced
                y = y_reduced
                noise = noise_reduced

            iteration = len(X)
            logging.info(
                f"Importing {iteration} existing datapoints. This could take a while..."
            )
            opt.tell(
                X,
                y,
                noise_vector=noise,
                gp_burnin=settings.get("gp_initial_burnin", gp_initial_burnin),
                gp_samples=settings.get("gp_initial_samples", gp_initial_samples),
                n_samples=settings.get("n_samples", 1),
                progress=True,
            )
            logging.info("Importing finished.")

    # 4. Main optimization loop:
    while True:
        logging.info("Starting iteration {}".format(iteration))
        result_every_n = settings.get("result_every", result_every)
        if (
            result_every_n > 0
            and iteration % result_every_n == 0
            and opt.gp.chain_ is not None
        ):
            result_object = create_result(Xi=X, yi=y, space=opt.space, models=[opt.gp])
            try:
                best_point, best_value = expected_ucb(result_object, alpha=0.0)
                best_point_dict = dict(zip(param_ranges.keys(), best_point))
                logging.info(f"Current optimum:\n{best_point_dict}")
                logging.info(f"Estimated value: {best_value}")
            except ValueError:
                logging.info(
                    "Computing current optimum was not successful. "
                    "This can happen in rare cases and running the "
                    "tuner again usually works."
                )
        plot_every_n = settings.get("plot_every", plot_every)
        if (
            plot_every_n > 0
            and iteration % plot_every_n == 0
            and opt.gp.chain_ is not None
        ):
            logging.getLogger("matplotlib.font_manager").disabled = True
            if opt.space.n_dims == 1:
                logging.warning(
                    "Plotting for only 1 parameter is not supported " "yet."
                )
            else:
                logging.debug("Starting to compute the next plot")
                result_object = create_result(
                    Xi=X, yi=y, space=opt.space, models=[opt.gp]
                )
                plt.style.use("dark_background")
                fig, ax = plt.subplots(
                    nrows=opt.space.n_dims,
                    ncols=opt.space.n_dims,
                    figsize=(3 * opt.space.n_dims, 3 * opt.space.n_dims),
                )
                fig.patch.set_facecolor("#36393f")
                for i in range(opt.space.n_dims):
                    for j in range(opt.space.n_dims):
                        ax[i, j].set_facecolor("#36393f")
                timestr = time.strftime("%Y%m%d-%H%M%S")
                plot_objective(
                    result_object, dimensions=list(param_ranges.keys()), fig=fig, ax=ax
                )
                plotpath = pathlib.Path(settings.get("plot_path", plot_path))
                plotpath.mkdir(parents=True, exist_ok=True)
                full_plotpath = plotpath / f"{timestr}-{iteration}.png"
                plt.savefig(
                    full_plotpath,
                    pad_inches=0.1,
                    dpi=300,
                    bbox_inches="tight",
                    facecolor="#36393f",
                )
                logging.info(f"Saving a plot to {full_plotpath}")
                plt.close(fig)
        point = opt.ask()
        point_dict = dict(zip(param_ranges.keys(), point))
        logging.info("Testing {}".format(point_dict))

        engine_json = prepare_engines_json(commands=commands, fixed_params=fixed_params)
        logging.debug(f"engines.json is prepared:\n{engine_json}")
        write_engines_json(engine_json, point_dict)
        logging.info("Start experiment")
        now = datetime.now()
        out_exp, out_exp_err = run_match(**settings)
        later = datetime.now()
        difference = (later - now).total_seconds()
        logging.info(f"Experiment finished ({difference}s elapsed)")
        logging.debug(f"Raw result:\n{out_exp}\n{out_exp_err}")

        score, error = parse_experiment_result(out_exp, **settings)
        logging.info("Got score: {} +- {}".format(score, error))
        logging.info("Updating model")
        while True:
            try:
                now = datetime.now()
                # We fetch kwargs manually here to avoid collisions:
                n_samples = settings.get("acq_function_samples", acq_function_samples)
                gp_burnin = settings.get("gp_burnin", gp_burnin)
                gp_samples = settings.get("gp_samples", gp_samples)
                if opt.gp.chain_ is None:
                    gp_burnin = settings.get("gp_initial_burnin", gp_initial_burnin)
                    gp_samples = settings.get("gp_initial_samples", gp_initial_samples)
                    opt.tell(
                        point,
                        score,
                        n_samples=n_samples,
                        gp_samples=gp_samples,
                        gp_burnin=gp_burnin,
                    )
                else:
                    opt.tell(
                        point,
                        score,
                        n_samples=n_samples,
                        gp_samples=gp_samples,
                        gp_burnin=gp_burnin,
                    )
                later = datetime.now()
                difference = (later - now).total_seconds()
                logging.info(f"GP sampling finished ({difference}s)")
                logging.debug(f"GP kernel: {opt.gp.kernel_}")
            except ValueError:
                logging.warning(
                    "Error encountered during fitting." "Trying to sample chain a bit."
                )
                opt.gp.sample(n_burnin=5, priors=opt.gp_priors)
            else:
                break
        X.append(point)
        y.append(score)
        noise.append(error)
        iteration = len(X)

        if os.name == "nt":
            np.savez_compressed(data_path, np.array(X), np.array(y), np.array(noise))
        else:
            with AtomicWriter(data_path, mode="wb", overwrite=True).open() as f:
                np.savez_compressed(f, np.array(X), np.array(y), np.array(noise))


if __name__ == "__main__":
    sys.exit(cli())  # pragma: no cover



from collections import namedtuple
from contextlib import contextmanager
from decimal import Decimal

import numpy as np
try:
    from sqlalchemy import create_engine
except ImportError:
    create_engine = None
from scipy.optimize import root_scalar
from scipy.special import expit

from tune.utils import parse_timecontrol

MatchResult = namedtuple("MatchResult", ["wins", "losses", "draws"])

ELO_CONSTANT = np.log(10) / 400.0


TC = namedtuple(
    "TimeControl",
    ["engine1_time", "engine1_increment", "engine2_time", "engine2_increment"],
)


class TimeControl(TC):
    @classmethod
    def from_strings(cls, engine1, engine2):
        tc1 = parse_timecontrol(engine1)
        tc2 = parse_timecontrol(engine2)
        inc1 = Decimal("0.0") if len(tc1) == 1 else tc1[1]
        inc2 = Decimal("0.0") if len(tc2) == 1 else tc2[1]
        return cls(
            engine1_time=Decimal(tc1[0]),
            engine1_increment=inc1,
            engine2_time=Decimal(tc2[0]),
            engine2_increment=inc2,
        )

    def to_strings(self):
        return (
            f"{self.engine1_time}+{self.engine1_increment}",
            f"{self.engine2_time}+{self.engine2_increment}",
        )


def get_session_maker(sessionmaker):
    @contextmanager
    def make_session():
        session = sessionmaker()
        try:
            yield session
            session.commit()
        except Exception:
            session.rollback()
            raise
        finally:
            session.close()

    return make_session


def create_sqlalchemy_engine(config):
    db_uri = (
        f"postgresql://{config['user']}:{config['password']}"
        f"@{config['host']}:{config['port']}/{config['dbname']}"
    )
    return create_engine(db_uri, pool_pre_ping=True)


def penta(ldw1, ldw2):
    x = np.fliplr(ldw1[:, None] * ldw2[None, :])
    sum_diagonals = np.vectorize(lambda offset: np.trace(x, offset=offset))
    ind = np.arange(-2, 3)[::-1]
    return sum_diagonals(ind)


def score_in_01(rates):
    k = len(rates)
    return np.sum(np.arange(k) * rates) / (k - 1)


def ldw_probabilities(elo, draw_elo, bias):
    pos = elo + bias
    w = expit(ELO_CONSTANT * (pos - draw_elo))
    l = expit(ELO_CONSTANT * (-pos - draw_elo))
    d = 1 - w - l
    return np.array([l, d, w])


def draw_rate_to_elo(draw_rate):
    return 400 * np.log10(2 / (1 - draw_rate) - 1)


def compute_probabilities_for_bias(elo, draw_elo, bias):
    ldw1 = ldw_probabilities(elo, draw_elo, bias)
    ldw2 = ldw_probabilities(elo, draw_elo, -bias)
    p = penta(ldw1, ldw2)
    return p


def compute_probabilities(elo, draw_elo, biases):
    result = np.empty((len(biases), 5))
    for i, b in enumerate(biases):
        result[i] = compute_probabilities_for_bias(elo, draw_elo, b)
    return result.mean(axis=0)


def elo_to_bayeselo(elo, draw_elo, biases):
    def func(bayeselo):
        return score_in_01(compute_probabilities(bayeselo, draw_elo, biases)) - expit(
            ELO_CONSTANT * elo
        )

    return root_scalar(func, method="brentq", bracket=(-1000, 1000)).root


def prior_from_drawrate(elo, draw_rate):
    biases = [-90, 200]
    draw_elo = draw_rate_to_elo(draw_rate)
    bayeselo = elo_to_bayeselo(elo, draw_elo=draw_elo, biases=biases)
    return compute_probabilities(bayeselo, draw_elo, biases)


def penta_to_score(draw_rate, counts, prior_games=10, prior_elo=0):
    prior = prior_games * prior_from_drawrate(prior_elo, draw_rate)
    probabilities = (counts + prior) / (counts.sum() + prior.sum())
    s01 = score_in_01(probabilities)
    score = (
        np.sqrt(2)
        * (s01 - 0.5)
        / (probabilities.dot(np.power(np.linspace(0, 2, 5), 2)) - 4 * s01 ** 2)
    )
    return score


def simple_penta_to_score(draw_rate, counts, prior_games=10, prior_elo=0):
    prior = prior_games * prior_from_drawrate(prior_elo, draw_rate)
    probabilities = (counts + prior) / (counts.sum() + prior.sum())
    s01 = score_in_01(probabilities)
    return s01 * 2 - 1



import json
import logging
import os
import re
import signal
import subprocess
import sys
import numpy as np
from time import sleep, time

try:
    from sqlalchemy.orm import sessionmaker
except ImportError:
    sessionmaker = None

from tune.db_workers.dbmodels import (
    Base,
    SqlJob,
    SqlUCIParam,
    SqlResult,
    SqlTimeControl,
    SqlTune,
)
from tune.io import InitStrings
from tune.db_workers.utils import (
    MatchResult,
    TimeControl,
    create_sqlalchemy_engine,
    get_session_maker,
)
from tune.utils import parse_timecontrol

CLIENT_VERSION = 2

__all__ = ["TuningClient"]


class TuningClient(object):
    def __init__(self, dbconfig_path, terminate_after=0, clientconfig=None, **kwargs):
        self.end_time = None
        if terminate_after != 0:
            start_time = time()
            self.end_time = start_time + terminate_after * 60
        self.logger = logging.getLogger("TuningClient")
        self.lc0_benchmark = None
        self.sf_benchmark = None
        signal.signal(signal.SIGINT, self.interrupt_handler)
        self.interrupt_pressed = False
        if os.path.isfile(dbconfig_path):
            with open(dbconfig_path, "r") as config_file:
                config = config_file.read().replace("\n", "")
                self.logger.debug(f"Reading DB config:\n{config}")
                self.connect_params = json.loads(config)
        else:
            raise ValueError(f"No config file found at provided path:\n{dbconfig_path}")

        self.engine = create_sqlalchemy_engine(self.connect_params)
        Base.metadata.create_all(self.engine)
        sm = sessionmaker(bind=self.engine)
        self.sessionmaker = get_session_maker(sm)

        self.client_config = None
        if clientconfig is not None:
            if os.path.isfile(clientconfig):
                with open(clientconfig, "r") as ccfile:
                    config = ccfile.read().replace("\n", "")
                    self.client_config = json.loads(config)
            else:
                raise ValueError(
                    f"Client configuration file not found:\n{clientconfig}"
                )

    def interrupt_handler(self, sig, frame):
        if self.interrupt_pressed:
            self.logger.info("Shutting down immediately.")
            sys.exit(0)
        self.interrupt_pressed = True
        self.logger.info(
            "Signal received. Shutting down after next match.\nPress a second time to terminate immediately."
        )

    def run_experiment(self, time_control, cutechess_options):
        try:
            os.remove("out.pgn")
        except OSError:
            pass

        st = [
            "cutechess-cli",
            "-concurrency",
            f"{cutechess_options['concurrency']}",
            "-engine",
            f"conf=engine1",
            f"tc={time_control.to_strings()[0]}",
            "-engine",
            "conf=engine2",
            f"tc={time_control.to_strings()[1]}",
            "-openings",
            f"file={cutechess_options['opening_path']}",
            "format=pgn",
            "order=random",
            "-draw",
            f"movenumber={cutechess_options['draw_movenumber']}",
            f"movecount={cutechess_options['draw_movecount']}",
            f"score={cutechess_options['draw_score']}",
            "-resign",
            f"movecount={cutechess_options['resign_movecount']}",
            f"score={cutechess_options['resign_score']}",
            "-rounds",
            f"{cutechess_options['rounds']}",
            "-repeat",
            "-games",
            "2",
            "-pgnout",
            "out.pgn",
        ]
        if "syzygy_path" in cutechess_options:
            st.insert(-2, "-tb")
            st.insert(-2, cutechess_options["syzygy_path"])
        out = subprocess.run(st, capture_output=True)
        return self.parse_experiment(out)

    def parse_experiment(self, results):
        string = results.stdout.decode("utf-8")
        err_string = results.stderr.decode("utf-8")
        error_occured = False
        if re.search("connection stalls", string):
            self.logger.error("Connection stalled during match. Aborting client.")
            error_occured = True
        elif re.search("Terminating process", string):
            self.logger.error("Engine was terminated. Aborting client.")
            error_occured = True
        if error_occured:
            self.logger.error(string)
            self.logger.error(err_string)
            sys.exit(1)
        lines = string.split("\n")
        self.logger.debug(f"Cutechess result string:\n{string}")
        last_output = lines[-4]
        result = re.findall(r"[0-9]\s-\s[0-9]\s-\s[0-9]", last_output)[0]
        w, l, d = [float(x) for x in re.findall("[0-9]", result)]
        return MatchResult(wins=w, losses=l, draws=d)

    def run_benchmark(self, config):
        def uci_to_cl(k, v):
            mapping = {
                "Threads": "--threads",
                "NNCacheSize": "--nncache",
                "Backend": "--backend",
                "BackendOptions": "--backend-opts",
                "MinibatchSize": "--minibatch-size",
                "MaxPrefetch": "--max-prefetch",
            }
            if k in mapping:
                return f"{mapping[k]}={v}"
            return None

        def cl_arguments(init_strings):
            cl_args = []
            for k, v in init_strings.items():
                arg = uci_to_cl(k, v)
                if arg is not None:
                    cl_args.append(arg)
            return cl_args

        self.logger.debug(
            f"Before benchmark engine 1:\n{config['engine'][0]['initStrings']}"
        )
        args = cl_arguments(InitStrings(config["engine"][0]["initStrings"]))
        self.logger.debug(f"Arguments for engine 1: {args}")
        path = os.path.join(os.path.curdir, "lc0")
        out = subprocess.run([path, "benchmark"] + args, capture_output=True)
        s = out.stdout.decode("utf-8")
        try:
            result = float(re.findall(r"([0-9\.]+)\snodes per second", s)[0])
        except IndexError:
            self.logger.error(f"Error while parsing engine1 benchmark:\n{s}")
            sys.exit(1)
        self.lc0_benchmark = result

        self.logger.debug(
            f"Before benchmark engine 2:\n{config['engine'][1]['initStrings']}"
        )
        num_threads = InitStrings(config["engine"][1]["initStrings"])["Threads"]
        path = os.path.join(os.path.curdir, "sf")
        out = subprocess.run(
            [path, "bench", "16", str(int(num_threads))], capture_output=True
        )
        # Stockfish outputs results as stderr:
        s = out.stderr.decode("utf-8")
        try:
            result = float(re.findall(r"Nodes/second\s+:\s([0-9]+)", s)[0])
        except IndexError:
            self.logger.error(f"Error while parsing engine2 benchmark:\n{s}")
            sys.exit(1)
        self.sf_benchmark = result

    def adjust_time_control(self, time_control, lc0_nodes, sf_nodes):
        lc0_ratio = lc0_nodes / self.lc0_benchmark
        sf_ratio = sf_nodes / self.sf_benchmark
        new_tc = TimeControl(
            engine1_time=float(time_control.engine1_time) * lc0_ratio,
            engine1_increment=float(time_control.engine1_increment) * lc0_ratio,
            engine2_time=float(time_control.engine2_time) * sf_ratio,
            engine2_increment=float(time_control.engine2_increment) * sf_ratio,
        )
        return new_tc

    @staticmethod
    def set_working_directories(job, engine_config):
        path = os.getcwd()
        engine_config[0]["workingDirectory"] = path
        engine_config[1]["workingDirectory"] = path
        exe1 = job.engine1_exe
        exe2 = job.engine2_exe
        if os.name == "nt":  # Windows needs .exe files to work correctly
            engine_config[0]["command"] = f"{exe1}.exe"
            engine_config[1]["command"] = f"{exe2}.exe"
        else:
            engine_config[0]["command"] = f"./{exe1}"
            engine_config[1]["command"] = f"./{exe2}"

    def pick_job(self, rows, mix=0.25):
        weights = []
        results = []
        min_samples = []
        for job, result in rows:
            weights.append(float(job.weight))
            results.append(result)
            min_samples.append(job.minimum_samplesize)
        """Pick a job based on weight and current load."""
        weights = np.array(weights)
        self.logger.debug(f"Job weights: {weights}")
        sample_size = np.array(
            [
                x.ww_count
                + x.wd_count
                + x.wl_count
                + x.dd_count
                + x.dl_count
                + x.ll_count
                for x in results
            ]
        )
        self.logger.debug(f"Sample sizes: {sample_size}")
        min_samples = np.array(min_samples)
        missing = np.maximum(min_samples - sample_size, 0.0)
        self.logger.debug(f"Missing samples: {missing}")
        if np.all(missing == 0.0):
            p = np.ones_like(weights) * weights
            p /= p.sum()
        else:
            uniform = np.ones_like(weights) / len(weights)
            p = missing * weights
            p /= p.sum()
            p = mix * uniform + (1 - mix) * p
        self.logger.debug(f"Resulting p={p}")
        rand_i = np.random.choice(len(results), p=p)
        result = results[rand_i]
        self.logger.debug(
            f"Picked result {rand_i} (job_id={result.job_id}, tc_id={result.tc_id})"
        )
        return rows[rand_i]

    def incorporate_config(self, config):
        admissible_uci = [
            "Threads",
            "Backend",
            "BackendOptions",
            "NNCacheSize",
            "MinibatchSize",
            "MaxPrefetch",
        ]
        engines = config["engine"]
        for i, e in enumerate(engines):
            engine_str = f"engine{i+1}"
            if engine_str in self.client_config:
                init_strings = InitStrings(e["initStrings"])
                for k, v in self.client_config[engine_str].items():
                    if k in admissible_uci:
                        init_strings[k] = v
        if "SyzygyPath" in self.client_config:
            path = self.client_config["SyzygyPath"]
            for e in engines:
                init_strings = InitStrings(e["initStrings"])
                init_strings["SyzygyPath"] = path
            config["cutechess"]["syzygy_path"] = path

    def run(self):
        while True:
            if self.interrupt_pressed:
                self.logger.info("Shutting down after receiving shutdown signal.")
                sys.exit(0)
            if self.end_time is not None and self.end_time < time():
                self.logger.info("Shutdown timer triggered. Closing")
                sys.exit(0)
            # 1. Check db for new job
            with self.sessionmaker() as session:
                rows = (
                    session.query(SqlJob, SqlResult)
                    .filter(SqlJob.id == SqlResult.job_id, SqlJob.active == True)
                    .all()
                )
                if len(rows) == 0:
                    sleep(30)  # TODO: maybe some sort of decay here
                    continue

                applicable_jobs = [
                    row for row in rows if row[0].minimum_version <= CLIENT_VERSION
                ]
                if len(applicable_jobs) < len(rows):
                    self.logger.warning(
                        "There are jobs which require a higher client version. Please update "
                        "the client as soon as possible!"
                    )
                    if len(applicable_jobs) == 0:
                        sleep(60)
                        continue
                job, sql_result = self.pick_job(applicable_jobs)

                # 2. Set up experiment
                # a) write engines.json
                config = json.loads(job.config)
                if self.client_config is not None:
                    self.incorporate_config(config)
                # TODO: remove: job_id = job["job_id"]
                self.logger.debug(f"Received config:\n{config}")
                engine_config = config["engine"]
                self.set_working_directories(job, engine_config)
                with open("engines.json", "w") as file:
                    json.dump(engine_config, file, sort_keys=True, indent=4)
                sleep(2)
                # b) Adjust time control:
                if self.lc0_benchmark is None:
                    self.logger.info(
                        "Running initial nodes/second benchmark to calibrate time controls."
                        "Ensure that your pc is idle to get a good reading."
                    )
                    self.run_benchmark(config)
                    self.logger.info(
                        f"Benchmark complete. Results: lc0: {self.lc0_benchmark} nps, sf: {self.sf_benchmark} nps"
                    )
                else:
                    self.logger.debug(
                        f"Initial benchmark results: lc0: {self.lc0_benchmark} nps, sf: {self.sf_benchmark} nps"
                    )
                orig_tc = sql_result.time_control.to_tuple()
                time_control = self.adjust_time_control(
                    orig_tc, float(job.engine1_nps), float(job.engine2_nps),
                )
                self.logger.debug(f"Adjusted time control from {orig_tc} to {time_control}")

                # 3. Run experiment (and block)
                self.logger.info(f"Running match with time control\n{time_control}")
                result = self.run_experiment(
                    time_control=time_control, cutechess_options=config["cutechess"]
                )
                self.logger.info(
                    f"Match result (WLD): {result.wins} - {result.losses} - {result.draws}"
                )
                # 5. Send results to database and lock it during access
                q = session.query(SqlResult).filter_by(
                        job_id=job.id, tc_id=sql_result.time_control.id
                    )
                if result.wins == 2:  # WW
                    q.update({"ww_count": SqlResult.ww_count + 1})
                elif result.wins == 1:
                    if result.draws == 1:  # WD
                        q.update({"wd_count": SqlResult.wd_count + 1})
                    else:  # WL
                        q.update({"wl_count": SqlResult.wl_count + 1})
                elif result.draws == 2:  # DD
                    q.update({"dd_count": SqlResult.dd_count + 1})
                elif result.draws == 1:  # DL
                    q.update({"dl_count": SqlResult.dl_count + 1})
                else:  # LL
                    q.update({"ll_count": SqlResult.ll_count + 1})
                self.logger.info("Uploaded match result to database.\n")



from skopt.plots import _format_scatter_plot_axes
from tune.utils import expected_ucb

import numpy as np
import matplotlib.pyplot as plt
from matplotlib.ticker import LogLocator


def _evenly_sample(dim, n_points):
    """Return `n_points` evenly spaced points from a Dimension.
    Parameters
    ----------
    dim : `Dimension`
        The Dimension to sample from.  Can be categorical; evenly-spaced
        category indices are chosen in order without replacement (result
        may be smaller than `n_points`).
    n_points : int
        The number of points to sample from `dim`.
    Returns
    -------
    xi : np.array
        The sampled points in the Dimension.  For Categorical
        dimensions, returns the index of the value in
        `dim.categories`.
    xi_transformed : np.array
        The transformed values of `xi`, for feeding to a model.
    """
    cats = np.array(getattr(dim, 'categories', []), dtype=object)
    if len(cats):  # Sample categoricals while maintaining order
        xi = np.linspace(0, len(cats) - 1, min(len(cats), n_points),
                         dtype=int)
        xi_transformed = dim.transform(cats[xi])
    else:
        bounds = dim.bounds
        # XXX use linspace(*bounds, n_points) after python2 support ends
        if dim.prior == "log-uniform":
            xi = np.logspace(np.log10(bounds[0]), np.log10(bounds[1]), n_points)
        else:
            xi = np.linspace(bounds[0], bounds[1], n_points)
        xi_transformed = dim.transform(xi)
    return xi, xi_transformed


def partial_dependence(space, model, i, j=None, sample_points=None,
                       n_samples=250, n_points=40, x_eval=None):
    """Calculate the partial dependence for dimensions `i` and `j` with
    respect to the objective value, as approximated by `model`.
    The partial dependence plot shows how the value of the dimensions
    `i` and `j` influence the `model` predictions after "averaging out"
    the influence of all other dimensions.
    When `x_eval` is not `None`, the given values are used instead of
    random samples. In this case, `n_samples` will be ignored.
    Parameters
    ----------
    space : `Space`
        The parameter space over which the minimization was performed.
    model
        Surrogate model for the objective function.
    i : int
        The first dimension for which to calculate the partial dependence.
    j : int, default=None
        The second dimension for which to calculate the partial dependence.
        To calculate the 1D partial dependence on `i` alone set `j=None`.
    sample_points : np.array, shape=(n_points, n_dims), default=None
        Only used when `x_eval=None`, i.e in case partial dependence should
        be calculated.
        Randomly sampled and transformed points to use when averaging
        the model function at each of the `n_points` when using partial
        dependence.
    n_samples : int, default=100
        Number of random samples to use for averaging the model function
        at each of the `n_points` when using partial dependence. Only used
        when `sample_points=None` and `x_eval=None`.
    n_points : int, default=40
        Number of points at which to evaluate the partial dependence
        along each dimension `i` and `j`.
    x_eval : list, default=None
        `x_eval` is a list of parameter values or None. In case `x_eval`
        is not None, the parsed dependence will be calculated using these
        values.
        Otherwise, random selected samples will be used.
    Returns
    -------
    For 1D partial dependence:
    xi : np.array
        The points at which the partial dependence was evaluated.
    yi : np.array
        The value of the model at each point `xi`.
    For 2D partial dependence:
    xi : np.array, shape=n_points
        The points at which the partial dependence was evaluated.
    yi : np.array, shape=n_points
        The points at which the partial dependence was evaluated.
    zi : np.array, shape=(n_points, n_points)
        The value of the model at each point `(xi, yi)`.
    For Categorical variables, the `xi` (and `yi` for 2D) returned are
    the indices of the variable in `Dimension.categories`.
    """
    # The idea is to step through one dimension, evaluating the model with
    # that dimension fixed and averaging either over random values or over
    # the given ones in x_val in all other dimensions.
    # (Or step through 2 dimensions when i and j are given.)
    # Categorical dimensions make this interesting, because they are one-
    # hot-encoded, so there is a one-to-many mapping of input dimensions
    # to transformed (model) dimensions.

    # If we haven't parsed an x_eval list we use random sampled values instead
    if x_eval is None and sample_points is None:
        sample_points = space.transform(space.rvs(n_samples=n_samples))
    elif sample_points is None:
        sample_points = space.transform([x_eval])

    # dim_locs[i] is the (column index of the) start of dim i in
    # sample_points.
    # This is usefull when we are using one hot encoding, i.e using
    # categorical values
    dim_locs = np.cumsum([0] + [d.transformed_size for d in space.dimensions])

    if j is None:
        # We sample evenly instead of randomly. This is necessary when using
        # categorical values
        xi, xi_transformed = _evenly_sample(space.dimensions[i], n_points)
        yi = []
        for x_ in xi_transformed:
            rvs_ = np.array(sample_points)  # copy
            # We replace the values in the dimension that we want to keep
            # fixed
            rvs_[:, dim_locs[i]:dim_locs[i + 1]] = x_
            # In case of `x_eval=None` rvs conists of random samples.
            # Calculating the mean of these samples is how partial dependence
            # is implemented.
            yi.append(np.mean(model.predict(rvs_)))

        return xi, yi

    else:
        xi, xi_transformed = _evenly_sample(space.dimensions[j], n_points)
        yi, yi_transformed = _evenly_sample(space.dimensions[i], n_points)

        zi = []
        for x_ in xi_transformed:
            row = []
            for y_ in yi_transformed:
                rvs_ = np.array(sample_points)  # copy
                rvs_[:, dim_locs[j]:dim_locs[j + 1]] = x_
                rvs_[:, dim_locs[i]:dim_locs[i + 1]] = y_
                row.append(np.mean(model.predict(rvs_)))
            zi.append(row)

        return xi, yi, np.array(zi).T


def plot_objective(
    result,
    levels=20,
    n_points=200,
    n_samples=30,
    size=3,
    zscale="linear",
    dimensions=None,
    n_random_restarts=100,
    alpha=0.25,
    colors=None,
    fig=None,
    ax=None,
):
    """Pairwise partial dependence plot of the objective function.
    The diagonal shows the partial dependence for dimension `i` with
    respect to the objective function. The off-diagonal shows the
    partial dependence for dimensions `i` and `j` with
    respect to the objective function. The objective function is
    approximated by `result.model.`
    Pairwise scatter plots of the points at which the objective
    function was directly evaluated are shown on the off-diagonal.
    A red point indicates the found minimum.
    Note: search spaces that contain `Categorical` dimensions are
          currently not supported by this function.
    Parameters
    ----------
    * `result` [`OptimizeResult`]
        The result for which to create the scatter plot matrix.
    * `levels` [int, default=10]
        Number of levels to draw on the contour plot, passed directly
        to `plt.contour()`.
    * `n_points` [int, default=40]
        Number of points at which to evaluate the partial dependence
        along each dimension.
    * `n_samples` [int, default=250]
        Number of random samples to use for averaging the model function
        at each of the `n_points`.
    * `size` [float, default=2]
        Height (in inches) of each facet.
    * `zscale` [str, default='linear']
        Scale to use for the z axis of the contour plots. Either 'linear'
        or 'log'.
    * `dimensions` [list of str, default=None] Labels of the dimension
        variables. `None` defaults to `space.dimensions[i].name`, or
        if also `None` to `['X_0', 'X_1', ..]`.
    * `n_random_restarts` [int, default=100]
        Number of restarts to try to find the global optimum.
    * `alpha` [float, default=0.5]
        Transparency of the sampled points.
    * `colors` [list of tuples, default=None]
        Colors to use for the optima.
    * `fig` [Matplotlib figure, default=None]
        Figure to use for plotting. If None, it will create one.
    * `ax` [k x k axes, default=None]
        Axes on which to plot the marginals. If None, it will create appropriate
        axes.
    Returns
    -------
    * `ax`: [`Axes`]:
        The matplotlib axes.
    """
    if colors is None:
        colors = plt.cm.get_cmap("Set3").colors
    space = result.space
    samples = np.asarray(result.x_iters)
    rvs_transformed = space.transform(space.rvs(n_samples=n_samples))

    if zscale == "log":
        locator = LogLocator()
    elif zscale == "linear":
        locator = None
    else:
        raise ValueError(
            "Valid values for zscale are 'linear' and 'log'," " not '%s'." % zscale
        )
    if fig is None:
        fig, ax = plt.subplots(
            space.n_dims,
            space.n_dims,
            figsize=(size * space.n_dims, size * space.n_dims),
        )

    fig.subplots_adjust(
        left=0.05, right=0.95, bottom=0.05, top=0.95, hspace=0.1, wspace=0.1
    )
    failures = 0
    while True:
        try:
            with result.models[-1].noise_set_to_zero():
                min_x = expected_ucb(
                    result, alpha=0.0, n_random_starts=n_random_restarts
                )[0]
                min_ucb = expected_ucb(result, n_random_starts=n_random_restarts)[0]
        except:
            failures += 1
            if failures == 10:
                break
            continue
        else:
            break

    for i in range(space.n_dims):
        for j in range(space.n_dims):
            if i == j:
                xi, yi = partial_dependence(
                    space,
                    result.models[-1],
                    i,
                    j=None,
                    sample_points=rvs_transformed,
                    n_points=n_points,
                )
                yi_min, yi_max = np.min(yi), np.max(yi)
                ax[i, i].plot(xi, yi, color=colors[1])
                if failures != 10:
                    ax[i, i].axvline(min_x[i], linestyle="--", color=colors[3], lw=1)
                    ax[i, i].axvline(min_ucb[i], linestyle="--", color=colors[5], lw=1)
                    ax[i, i].text(
                        min_x[i],
                        yi_min + 0.9 * (yi_max - yi_min),
                        f"{np.around(min_x[i], 4)}",
                        color=colors[3],
                    )
                    ax[i, i].text(
                        min_ucb[i],
                        yi_min + 0.7 * (yi_max - yi_min),
                        f"{np.around(min_ucb[i], 4)}",
                        color=colors[5],
                    )

            # lower triangle
            elif i > j:
                xi, yi, zi = partial_dependence(
                    space, result.models[-1], i, j, rvs_transformed, n_points
                )
                ax[i, j].contourf(xi, yi, zi, levels, locator=locator, cmap="viridis_r")
                ax[i, j].scatter(
                    samples[:, j], samples[:, i], c="k", s=10, lw=0.0, alpha=alpha
                )
                if failures != 10:
                    ax[i, j].scatter(min_x[j], min_x[i], c=["r"], s=20, lw=0.0)
                    ax[i, j].scatter(
                        min_ucb[j], min_ucb[i], c=["xkcd:orange"], s=20, lw=0.0
                    )

    return _format_scatter_plot_axes(
        ax, space, ylabel="Partial dependence", dim_labels=dimensions
    )



import re
import pathlib
import subprocess

import numpy as np
from scipy.stats import dirichlet
from skopt.space import Categorical, Integer, Real

from tune.utils import TimeControl


__all__ = ["run_match", "parse_experiment_result", "reduce_ranges"]


def parse_experiment_result(
    outstr, prior_counts=None, n_dirichlet_samples=1000000, **kwargs
):
    """Parse cutechess-cli result output to extract mean score and error.

    Here we use a simple pentanomial model to exploit paired openings.
    We distinguish the outcomes WW, WD, WL/DD, LD and LL and apply the
    following scoring (note, that the optimizer always minimizes the score):

    +------+------+-------+-----+-----+
    | WW   | WD   | WL/DD | LD  | LL  |
    +======+======+=======+=====+=====+
    | -1.0 | -0.5 | 0.0   | 0.5 | 1.0 |
    +------+------+-------+-----+-----+

    Note: It is important that the match output was produced using
    cutechess-cli using paired openings, otherwise the returned score is
    useless.

    Parameters
    ----------
    output : string (utf-8)
        Match output of cutechess-cli. It assumes the output was coming from
        a head-to-head match with paired openings.
    prior_counts : list-like float or int, default=None
        Pseudo counts to use for WW, WD, WL/DD, LD and LL in the
        pentanomial model.
    n_dirichlet_samples : int, default = 1 000 000
        Number of samples to draw from the Dirichlet distribution in order to
        estimate the standard error of the score.
    Returns
    -------
    score : float (in [-1, 1])
        Expected (negative) score of the first player (the lower the stronger)
    error : float
        Estimated standard error of the score. Estimated by repeated draws
        from a Dirichlet distribution.
    """
    wdl_strings = re.findall(r"[0-9]+\s-\s[0-9]+\s-\s[0-9]+", outstr)
    array = np.array(
        [np.array([int(y) for y in re.findall(r"[0-9]+", x)]) for x in wdl_strings]
    )
    diffs = np.diff(array, axis=0, prepend=np.array([[0, 0, 0]]))

    counts = {"WW": 0, "WD": 0, "WL/DD": 0, "LD": 0, "LL": 0}
    for i in range(0, len(diffs) - 1, 2):
        match = diffs[i] + diffs[i + 1]
        if match[0] == 2:
            counts["WW"] += 1
        elif match[0] == 1:
            if match[1] == 1:
                counts["WL/DD"] += 1
            else:
                counts["WD"] += 1
        elif match[1] == 1:
            counts["LD"] += 1
        elif match[2] == 2:
            counts["WL/DD"] += 1
        else:
            counts["LL"] += 1
    counts_array = np.array(list(counts.values()))
    if prior_counts is None:
        prior_counts = np.array([0.14, 0.19, 0.34, 0.19, 0.14]) * 2.5
    elif len(prior_counts) != 5:
        raise ValueError("Argument prior_counts should contain 5 elements.")
    dist = dirichlet(alpha=counts_array + prior_counts)
    scores = [-1.0, -0.5, 0.0, 0.5, 1.0]
    score = dist.mean().dot(scores)
    error = dist.rvs(n_dirichlet_samples).dot(scores).var()
    return score, error


def _construct_engine_conf(id, engine_npm=None, engine_tc=None):
    result = ["-engine", f"conf=engine{id}"]
    if engine_npm is not None:
        result.extend(("tc=inf", f"nodes={engine_npm}"))
        return result
    if isinstance(engine_tc, str):
        engine_tc = TimeControl.from_string(engine_tc)
    result.append(f"tc={str(engine_tc)}")
    return result


def run_match(
    rounds=5,
    engine1_tc=None,
    engine2_tc=None,
    engine1_npm=None,
    engine2_npm=None,
    opening_file=None,
    adjudicate_draws=True,
    draw_movenumber=1,
    draw_movecount=10,
    draw_score=8,
    adjudicate_resign=True,
    resign_movecount=3,
    resign_score=550,
    adjudicate_tb=False,
    tb_path=None,
    concurrency=1,
    output_as_string=True,
    **kwargs,
):
    """Run a cutechess-cli match of two engines with paired random openings.

    Parameters
    ----------
    rounds : int, default=1
        Number of rounds to play in the match (each round consists of 2 games).
    engine1_tc : str or TimeControl object, default=None
        Time control to use for the first engine. If str, it can be a
        non-increment time control like "10" (10 seconds) or an increment
        time control like "5+1.5" (5 seconds total with 1.5 seconds increment).
        If None, it is assumed that engine1_npm is provided.
    engine2_tc : str or TimeControl object, default=None
        See engine1_tc.
    engine1_npm : str or int, default=None
        Number of nodes per move the engine is allowed to search.
        If None, it is assumed that engine1_tc is provided.
    engine2_npm : str or int, default=None
        See engine1_npm.
    opening_file : str, default=None
        Path to the file containing the openings. Can be .epd or .pgn.
        Make sure that the file explicitly has the .epd or .pgn suffix, as it
        is used to detect the format.
    adjudicate_draws : bool, default=True
        Specify, if cutechess-cli is allowed to adjudicate draws, if the
        scores of both engines drop below draw_score for draw_movecount number
        of moves. Only kicks in after draw_movenumber moves have been played.
    draw_movenumber : int, default=1
        Number of moves to play after the opening, before draw adjudication is
        allowed.
    draw_movecount : int, default=10
        Number of moves below the threshold draw_score, without captures and
        pawn moves, before the game is adjudicated as draw.
    draw_score : int, default=8
        Score threshold of the engines in centipawns. If the score of both
        engines drops below this value for draw_movecount consecutive moves,
        and there are no captures and pawn moves, the game is adjudicated as
        draw.
    adjudicate_resign : bool, default=True
        Specify, if cutechess-cli is allowed to adjudicate wins/losses based on
        the engine scores. If one engine’s score drops below -resign_score for
        resign_movecount many moves, the game is considered a loss for this
        engine.
    resign_movecount : int, default=3
        Number of consecutive moves one engine has to output a score below
        the resign_score threshold for the game to be considered a loss for this
        engine.
    resign_score : int, default=550
        Resign score threshold in centipawns. The score of the engine has to
        stay below -resign_score for at least resign_movecount moves for it to
        be adjudicated as a loss.
    adjudicate_tb : bool, default=False
        Allow cutechess-cli to adjudicate games based on Syzygy tablebases.
        If true, tb_path has to be set.
    tb_path : str, default=None
        Path to the
    concurrency : int, default=1
        Number of games to run in parallel. Be careful when running time control
        games, since the engines can negatively impact each other when running
        in parallel.
    output_as_string : bool, default=True
        If True, only return the cutechess-cli output as string.
        If False, the complete subprocess.CompletedProcess object will be
        returned for debugging purposes.

    Returns
    -------
    out : str or subprocess.CompletedProcess object
        Results of the cutechess-cli match as string.
        If output_as_string was set to False, returns the
        CompletedProcess object for debugging purposes.
    """
    string_array = ["cutechess-cli"]
    string_array.extend(("-concurrency", str(concurrency)))

    if (engine1_npm is None and engine1_tc is None) or (
        engine2_npm is None and engine2_tc is None
    ):
        raise ValueError("A valid time control or nodes configuration is required.")
    string_array.extend(_construct_engine_conf(1, engine1_npm, engine1_tc))
    string_array.extend(_construct_engine_conf(2, engine2_npm, engine2_tc))

    if opening_file is None:
        raise ValueError("Providing an opening file is required.")
    opening_path = pathlib.Path(opening_file)
    if not opening_path.exists():
        raise FileNotFoundError(
            f"Opening file the following path was not found: {opening_path}"
        )
    opening_format = opening_path.suffix
    if opening_format not in {".epd", ".pgn"}:
        raise ValueError(
            "Unable to determine opening format. "
            "Make sure to add .epd or .pgn to your filename."
        )
    string_array.extend(
        (
            "-openings",
            f"file={str(opening_path)}",
            f"format={opening_format[1:]}",
            "order=random",
        )
    )

    if adjudicate_draws:
        string_array.extend(
            (
                "-draw",
                f"movenumber={draw_movenumber}",
                f"movecount={draw_movecount}",
                f"score={draw_score}",
            )
        )
    if adjudicate_resign:
        string_array.extend(
            ("-resign", f"movecount={resign_movecount}", f"score={resign_score}")
        )
    if adjudicate_tb:
        if tb_path is None:
            raise ValueError("No path to tablebases provided.")
        tb_path_object = pathlib.Path(tb_path)
        if not tb_path_object.exists():
            raise FileNotFoundError(
                f"No folder found at the following path: {str(tb_path_object)}"
            )
        string_array.extend(("-tb", str(tb_path_object)))

    string_array.extend(("-rounds", f"{rounds}"))
    string_array.extend(("-games", "2"))
    string_array.append("-repeat")
    string_array.append("-recover")
    string_array.extend(("-pgnout", "out.pgn"))

    out = subprocess.run(string_array, capture_output=True)
    if output_as_string:
        return out.stdout.decode("utf-8"), out.stderr.decode("utf-8")
    return out


def reduce_ranges(X, y, noise, space):
    X_new = []
    y_new = []
    noise_new = []
    reduction_needed = False
    for row, yval, nval in zip(X, y, noise):
        include_row = True
        for dim, value in zip(space.dimensions, row):
            if isinstance(dim, Integer) or isinstance(dim, Real):
                lb, ub = dim.bounds
                if value < lb or value > ub:
                    include_row = False
            elif isinstance(dim, Categorical):
                if value not in dim.bounds:
                    include_row = False
            else:
                raise ValueError(f"Parameter type {type(dim)} unknown.")
        if include_row:
            X_new.append(row)
            y_new.append(yval)
            noise_new.append(nval)
        else:
            reduction_needed = True
    return reduction_needed, X_new, y_new, noise_new



"""Top-level package for Chess Tuning Tools."""

__author__ = """Karlson Pfannschmidt"""
__email__ = "kiudee@mail.upb.de"
__version__ = "0.5.0-beta.1"

from tune.utils import expected_ucb
from tune.db_workers import TuningClient, TuningServer



from ast import literal_eval
from collections.abc import MutableMapping
import json
from pathlib import Path
import re
import subprocess
import sys

import skopt.space as skspace
from skopt.space.space import check_dimension


__all__ = [
    "InitStrings",
    "uci_tuple",
]


# TODO: Backup file to restore it, should there be an error
def uci_tuple(uci_string):
    try:
        name, value = re.findall(r"name\s+(\S.*?)\s+value\s+(.*?)\s*$", uci_string)[0]
    except IndexError:
        print(f"Error parsing UCI tuples:\n{uci_string}")
        sys.exit(1)
    try:
        tmp = float(value)
    except ValueError:
        tmp = value
    return name, tmp


def set_option(name, value):
    return f"setoption name {name} value {value}"


class InitStrings(MutableMapping):
    def __init__(self, init_strings):
        self._init_strings = init_strings

    def __len__(self):
        return len(self._init_strings)

    def __getitem__(self, key):
        for s in self._init_strings:
            name, value = uci_tuple(s)
            if key == name:
                return value
        raise KeyError(key)

    def __setitem__(self, key, value):
        for i, s in enumerate(self._init_strings):
            name, _ = uci_tuple(s)
            if key == name:
                self._init_strings[i] = set_option(key, value)
                return
        self._init_strings.append(set_option(key, value))

    def __delitem__(self, key):
        elem = -1
        for i, s in enumerate(self._init_strings):
            name, _ = uci_tuple(s)
            if key == name:
                elem = i
                break
        if elem != -1:
            del self._init_strings[i]
        else:
            raise KeyError(key)

    def __contains__(self, key):
        for s in self._init_strings:
            name, _ = uci_tuple(s)
            if key == name:
                return True
        return False

    def __iter__(self):
        for s in self._init_strings:
            name, _ = uci_tuple(s)
            yield name

    def __repr__(self):
        return repr(self._init_strings)


def make_numeric(s):
    try:
        return int(s)
    except ValueError:
        try:
            return float(s)
        except ValueError:
            return s


def parse_ranges(s):
    if isinstance(s, str):
        j = json.loads(s)
    else:
        j = s

    dimensions = []
    for i, s in enumerate(j.values()):
        # First check, if the string is a list/tuple or a function call:
        param_str = re.findall(r"(\w+)\(", s)
        if len(param_str) > 0:  # Function
            args, kwargs = [], dict()
            # TODO: this split does not always work (example Categorical(["a", "b", "c"]))
            prior_param_strings = re.findall(r"\((.*?)\)", s)[0].split(",")
            for arg_string in prior_param_strings:
                # Check if arg or kwarg:
                if "=" in arg_string:  # kwarg:
                    # trim all remaining whitespace:
                    arg_string = "".join(arg_string.split())

                    key, val = arg_string.split("=")
                    kwargs[key] = make_numeric(val)
                elif "[" in arg_string or "(" in arg_string:
                    args.append(literal_eval(arg_string))
                else:  # args:
                    val = make_numeric(arg_string)
                    args.append(val)
            if hasattr(skspace, param_str[0]):
                dim = getattr(skspace, param_str[0])(*args, **kwargs)
            else:
                raise ValueError("Dimension {} does not exist.".format(param_str))
            dimensions.append(dim)
        else:  # Tuple or list
            # We assume that the contents of the collection should be used as is and construct a python list/tuple
            # skopt.space.check_dimension will be used for validation
            parsed = literal_eval(s)
            if isinstance(parsed, (tuple, list)):
                dimensions.append(check_dimension(parsed))
            else:
                raise ValueError(
                    "Dimension {} is not valid. Make sure you pass a Dimension, tuple or list.".format(
                        param_str
                    )
                )

    return dict(zip(j.keys(), dimensions))


def load_tuning_config(json_dict):
    """ Load the provided tuning configuration and split it up.

    Parameters
    ----------
    json_dict : dict
        Dictionary containing the engines, their fixed parameters, the tunable
        parameter ranges and other settings used during tuning.

    Returns
    -------
    json_dict : dict
        Remaining settings after the engine configuration and the ranges have
        been stripped off
    commands : list of strings
    fixed_params : list of dicts
        UCI parameters to be set for the engines.
    param_ranges : dict
        UCI parameters of the first engine which are to be optimized during tuning.
        The values correspond to skopt.space dimension definitions.
    """
    commands = []
    fixed_params = []
    if "engines" not in json_dict:
        raise ValueError("Tuning config does not contain engines.")
    engines = json_dict["engines"]
    for e in engines:
        if "command" not in e:
            raise ValueError("Tuning config contains an engine without command.")
        commands.append(e["command"])
        if "fixed_parameters" not in e:
            fixed_params.append(dict())
        else:
            fixed_params.append(e["fixed_parameters"])
    del json_dict["engines"]
    if "parameter_ranges" not in json_dict:
        raise ValueError("There are no parameter ranges defined in the config file.")
    param_ranges = parse_ranges(json_dict["parameter_ranges"])
    del json_dict["parameter_ranges"]
    # All remaining settings will be returned as is:
    return json_dict, commands, fixed_params, param_ranges


def prepare_engines_json(commands, fixed_params):
    result_list = [
        {"command": c, "name": f"engine{i+1}", "initStrings": [], "protocol": "uci"}
        for i, c in enumerate(commands)
    ]
    for r, fp in zip(result_list, fixed_params):
        uci = InitStrings(r["initStrings"])
        uci.update(fp)
    return result_list


def write_engines_json(engine_json, point_dict):
    engine = engine_json[0]
    initstr = InitStrings(engine["initStrings"])
    initstr.update(point_dict)
    with open(Path() / "engines.json", "w") as file:
        json.dump(engine_json, file, sort_keys=True, indent=4)


class Engines(object):
    def __init__(self, parameters, engine1_fixed, engine2_fixed):
        pass

    def set_params(self, values):
        """Set tunable parameters to new values.

        Parameters
        ----------
        values :

        """
        pass

    def write_json(self, path=None):
        """Write the engines.json file

        Parameters
        ----------
        path :

        """
        pass



from datetime import datetime

__all__ = [
    "SqlTune",
    "SqlJob",
    "SqlUCIParam",
    "SqlTimeControl",
    "SqlResult",
]
try:
    from sqlalchemy import (
        Boolean,
        Column,
        DateTime,
        ForeignKey,
        Integer,
        JSON,
        Numeric,
        PrimaryKeyConstraint,
        String,
        Table,
    )
    from sqlalchemy.orm import relationship

    from tune.db_workers.dbmodels import Base
    from tune.db_workers.utils import TimeControl

    SCHEMA = "new"

    class SqlTune(Base):
        __tablename__ = "tunes"
        __table_args__ = {"schema": SCHEMA}
        id = Column(Integer, primary_key=True)
        timestamp = Column(DateTime, default=datetime.utcnow, nullable=False)
        weight = Column(Numeric, default=1.0, nullable=False)
        description = Column(String(250), nullable=True)

        jobs = relationship("SqlJob", back_populates="tune", cascade="all")

        def __repr__(self):
            return f"<Tune (id={self.id}, timestamp={self.timestamp}, weight={self.weight})>"

    job_tc_table = Table(
        "jobstotimes",
        Base.metadata,
        Column("job_id", Integer, ForeignKey(f"{SCHEMA}.jobs.id")),
        Column("tc_id", Integer, ForeignKey(f"{SCHEMA}.timecontrols.id")),
        schema=SCHEMA,
    )

    class SqlJob(Base):
        __tablename__ = "jobs"
        __table_args__ = {"schema": SCHEMA}
        id = Column(Integer, primary_key=True)
        timestamp = Column(DateTime, default=datetime.utcnow, nullable=False)
        active = Column(Boolean, default=True, nullable=False)
        weight = Column(Numeric, default=1.0, nullable=False)
        minimum_version = Column(Integer, default=1, nullable=False)
        maximum_version = Column(Integer, nullable=True, default=None)
        minimum_samplesize = Column(Integer, nullable=False, default=16)
        config = Column(JSON)
        engine1_exe = Column(String(100), nullable=False, default="lc0")
        engine1_nps = Column(Numeric, nullable=False)
        engine2_exe = Column(String(100), nullable=False, default="sf")
        engine2_nps = Column(Numeric, nullable=False)

        tune_id = Column(Integer, ForeignKey(f"{SCHEMA}.tunes.id"))
        tune = relationship("SqlTune", back_populates="jobs")
        params = relationship("SqlUCIParam", back_populates="job", cascade="all")
        time_controls = relationship("SqlTimeControl", secondary=job_tc_table)
        results = relationship("SqlResult", back_populates="job")

        def __repr__(self):
            return (
                f"<Job (id={self.id}, timestamp={self.timestamp}, active={self.active},"
                f" weight={self.weight}, engine1_exe={self.engine1_exe}, engine1_nps={self.engine1_nps},"
                f" engine2_exe={self.engine2_exe}, engine2_nps={self.engine2_nps}, config={self.config})>"
            )

    class SqlUCIParam(Base):
        __tablename__ = "params"
        __table_args__ = (
            PrimaryKeyConstraint("key", "job_id", name="param_pk"),
            {"schema": SCHEMA},
        )
        key = Column(String(100))
        value = Column(String(250), nullable=False)
        job_id = Column(Integer, ForeignKey(f"{SCHEMA}.jobs.id"))
        job = relationship("SqlJob", back_populates="params")

        def __repr__(self):
            return f"<UCI (key={self.key}, value={self.value}, job_id={self.job_id})>"

    class SqlTimeControl(Base):
        __tablename__ = "timecontrols"
        __table_args__ = ({"schema": SCHEMA},)
        id = Column(Integer, primary_key=True)
        engine1_time = Column(Numeric, nullable=False)
        engine1_increment = Column(Numeric, nullable=True, default=None)
        engine2_time = Column(Numeric, nullable=False)
        engine2_increment = Column(Numeric, nullable=True, default=None)
        draw_rate = Column(Numeric, nullable=False, default=0.33)

        def __repr__(self):
            engine1_inc = (
                "" if self.engine1_increment is None else f"+{self.engine1_increment}"
            )
            engine2_inc = (
                "" if self.engine2_increment is None else f"+{self.engine2_increment}"
            )
            return (
                f"<TC (id={self.id}, engine1={self.engine1_time}{engine1_inc},"
                f" engine2={self.engine2_time}{engine2_inc}, draw_rate={self.draw_rate})>"
            )

        def to_tuple(self):
            return TimeControl(
                engine1_time=self.engine1_time,
                engine1_increment=self.engine1_increment,
                engine2_time=self.engine2_time,
                engine2_increment=self.engine2_increment,
            )

    class SqlResult(Base):
        __tablename__ = "results"
        __table_args__ = ({"schema": SCHEMA},)
        job_id = Column(Integer, ForeignKey(f"{SCHEMA}.jobs.id"), primary_key=True)
        tc_id = Column(
            Integer, ForeignKey(f"{SCHEMA}.timecontrols.id"), primary_key=True
        )
        ww_count = Column(Integer, default=0)
        wl_count = Column(Integer, default=0)
        wd_count = Column(Integer, default=0)
        dd_count = Column(Integer, default=0)
        dl_count = Column(Integer, default=0)
        ll_count = Column(Integer, default=0)
        job = relationship("SqlJob", back_populates="results")
        time_control = relationship("SqlTimeControl")

        def __repr__(self):
            return (
                f"<Result (job_id={self.job_id}, tc_id={self.tc_id},"
                f" ww={self.ww_count}, wd={self.wd_count},"
                f" wl={self.wl_count}, dd={self.dd_count},"
                f" dl={self.dl_count}, ll={self.ll_count})>"
            )


except ImportError:
    SqlTune = None
    SqlJob = None
    SqlUCIParam = None
    SqlTimeControl = None
    SqlResult = None



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['tune', 'tune.db_workers', 'tune.db_workers.dbmodels']

package_data = \
{'': ['*']}

install_requires = \
['Click>=7.1.2,<8.0.0',
 'atomicwrites>=1.4.0,<2.0.0',
 'bask>=0.7.2,<0.8.0',
 'emcee>=3.0.2,<4.0.0',
 'numpy>=1.19.1,<2.0.0',
 'scikit-optimize>=0.7.4,<0.8.0',
 'scipy>=1.5.2,<2.0.0']

extras_require = \
{'dist': ['joblib>=0.16.0,<0.17.0',
          'psycopg2>=2.8.5,<3.0.0',
          'sqlalchemy>=1.3.18,<2.0.0',
          'pandas>=1.1.0,<2.0.0']}

entry_points = \
{'console_scripts': ['tune = tune.cli:cli']}

setup_kwargs = {
    'name': 'chess-tuning-tools',
    'version': '0.5.0b1',
    'description': 'A collection of tools for local and distributed tuning of chess engines.',
    'long_description': '==================\nChess Tuning Tools\n==================\n\n\n.. image:: https://img.shields.io/pypi/v/chess-tuning-tools.svg\n        :target: https://pypi.python.org/pypi/chess-tuning-tools\n\n.. image:: https://img.shields.io/travis/kiudee/chess-tuning-tools.svg\n        :target: https://travis-ci.org/kiudee/chess-tuning-tools\n\n.. image:: https://readthedocs.org/projects/chess-tuning-tools/badge/?version=latest\n        :target: https://chess-tuning-tools.readthedocs.io/en/latest/?badge=latest\n        :alt: Documentation Status\n\n\n\n\nA collection of tools for local and distributed tuning of chess engines.\n\n\n* Free software: Apache Software License 2.0\n* Documentation: https://chess-tuning-tools.readthedocs.io.\n\n\nFeatures\n--------\n\n* Optimization of chess engines using state-of-the-art Bayesian optimization.\n* Support for automatic visualization of the optimization landscape.\n* Scoring matches using the pentanomial model for paired openings.\n\nQuick Start\n-----------\n\nIn order to be able to start the tuning, first create a python\nenvironment and install chess-tuning-tools by typing::\n\n   pip install chess-tuning-tools\n\nFurthermore, you need to have `cutechess-cli <https://github.com/cutechess/cutechess>`_\nin the path. The tuner will use it to run matches.\n\nTo execute the local tuner, simply run::\n\n   tune local -c tuning_config.json\n\nTake a look at the `usage instructions`_ and the `example configurations`_ to\nlearn how to set up the ``tuning_config.json`` file.\n\n\nDistributed tuning\n------------------\n\nThe distributed tuning framework is currently not actively supported.\nTo be able to run the tuning client, you need the following directory structure::\n\n   folder/\n   |---- networks/\n   |     |---- networkid\n   |---- openings/\n   |     |---- ...\n   |     |---- openings.pgn\n   |     |---- ...\n   |---- dbconfig.json\n   |---- engine1[.exe]\n   |---- engine2[.exe]\n\nFinally, the tuning client can be started as follows::\n\n   cd path/to/folder\n   tune run-client dbconfig.json\n\nThe client can be terminated gracefully by inputting ctrl-c once or terminated\nimmediately by sending it twice.\n\nYou will also need to run a PostgreSQL database, which the server will use to\npost jobs for the clients to fetch and the clients to report their results to.\n\nCredits\n-------\n\nThis package was created with Cookiecutter_ and the `audreyr/cookiecutter-pypackage`_ project template.\n\n.. _Cookiecutter: https://github.com/audreyr/cookiecutter\n.. _`audreyr/cookiecutter-pypackage`: https://github.com/audreyr/cookiecutter-pypackage\n.. _example configurations: https://github.com/kiudee/chess-tuning-tools/tree/master/examples\n.. _usage instructions: https://chess-tuning-tools.readthedocs.io/en/latest/usage.html\n',
    'author': 'Karlson Pfannschmidt',
    'author_email': 'kiudee@mail.upb.de',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/kiudee/chess-tuning-tools',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
